// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package models

import (
	"bytes"
	"fmt"
	"net/http"
	"net/http/httptest"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicegroupchangetype"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

var _ = Describe("API Group Model validation tests", func() {
	orgKeyObject, _ := orgkeylib.Encode("enki", 1337)
	orgKey := orgKeyObject.String()

	// Sample body definition used by many tests
	rawJSON := `{"changes": ["DG_LIVE_QUERY_CHANGE", "DG_POLICY_CHANGE"], "live_query_time":1672531200000}`

	// ----- URL Bind testing --------------------
	// NOTE: These handle supplied url API content validation, not value validation

	// NOTE: Same effect as supplying empty string orgKey
	It("Ensure we require orgKey", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawJSON))
		context.Params = []gin.Param{
			{Key: "deviceGroupId", Value: "1337"},
		}
		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindUri(apiGroupModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceGroupURI.OrgKey' Error:Field validation for 'OrgKey' failed on the 'required' tag"))
	})

	// NOTE: Same effect as supplying empty string deviceGroupId
	It("Ensure we require a deviceGroupId", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawJSON))
		context.Params = []gin.Param{
			{Key: "orgKey", Value: orgKey},
		}
		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindUri(apiGroupModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceGroupURI.Id' Error:Field validation for 'Id' failed on the 'required' tag"))
	})

	It("Ensure we reject a non-int deviceGroupId", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawJSON))
		context.Params = []gin.Param{
			{Key: "orgKey", Value: orgKey},
			{Key: "deviceGroupId", Value: "BOGUS"},
		}
		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindUri(apiGroupModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("strconv.ParseInt: parsing \"BOGUS\": invalid syntax"))
	})

	// ----- Body Bind testing --------------------
	// NOTE: These handle supplied body API content validation, not value validation

	It("Ensure we can parse a valid device structure", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawJSON))

		apiDeviceGroup := APIDeviceGroup{}
		err := context.ShouldBindJSON(apiDeviceGroup.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiDeviceGroup.Changes)).To(BeEquivalentTo(2))
		Expect(*apiDeviceGroup.LiveQueryTime).To(BeEquivalentTo(1672531200000))
	})

	It("Ensure we reject when empty string used for body", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(""))

		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindJSON(apiGroupModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("EOF")) // NOTE: best we can do with gomega
	})

	It("Ensure we accept empty JSON body", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString("{}"))

		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindJSON(apiGroupModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiGroupModel.Changes)).To(BeEquivalentTo(0))
		Expect(apiGroupModel.LiveQueryTime).To(BeNil())
	})

	// -----

	It("Ensure we reject non-list changes", func() {
		rawGroupString := `{"changes":"DG_LIVE_QUERY_CHANGE"}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiGroupModel := APIDeviceGroup{}

		err := context.ShouldBindJSON(apiGroupModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIDeviceGroupBody.changes of type []string"))
	})

	It("Ensure we accept empty changes", func() {
		rawGroupString := `{"changes":[]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindJSON(apiGroupModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiGroupModel.Changes)).To(BeEquivalentTo(0))
	})

	badJSON := `{"changes": ["DG_LIVE_QUERY_CHANGE", "DG_POLICY_CHANGE", "DG_POLICY_CHANGE"]}`

	It("Ensure we reject excess changes", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(badJSON))

		apiGroupModel := APIDeviceGroup{}
		err := context.ShouldBindJSON(apiGroupModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceGroupBody.Changes' Error:Field validation for 'Changes' failed on the 'max' tag"))
	})

	// ----- DB Conversion testing --------------------
	// NOTE: These validate DB structure conversion checks

	useAPI := APIDeviceGroupURI{Id: 1337, OrgKey: orgKey}

	It("Ensure we convert a valid group device structure", func() {
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: useAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes: []string{devicegroupchangetype.DG_POLICY_CHANGE.String(), devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()},
			},
		}
		group, err := apiDeviceGroup.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(group).ToNot(BeNil())
		Expect(group.Id).To(BeEquivalentTo(1337))
		Expect(group.OrgKey).To(BeEquivalentTo(orgKey))
		Expect(group.LastPolicyChangeTime).ToNot(BeNil())
		Expect(group.LiveQueryChangeTime).ToNot(BeNil())
		Expect(group.GetEnv()).To(BeEquivalentTo("enki"))
	})

	badAPI := APIDeviceGroupURI{Id: 1337, OrgKey: "BOGUS456"}

	It("Ensure we catch an invalid orgkey", func() {
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: badAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes: []string{devicegroupchangetype.DG_POLICY_CHANGE.String(), devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()},
			},
		}
		group, err := apiDeviceGroup.ConvertToDBObject()
		Expect(group).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("error decoding org key BOGUS456: alphabet used for hash was different"))
	})

	badAPI2 := APIDeviceGroupURI{Id: 1337, OrgKey: "       "}

	It("Ensure we catch an emptyspace orgkey", func() {
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: badAPI2,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes: []string{devicegroupchangetype.DG_POLICY_CHANGE.String(), devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()},
			},
		}
		group, err := apiDeviceGroup.ConvertToDBObject()
		Expect(group).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("org key is required"))
	})

	It("Ensure we reject an invalid DeviceChangeType for Changes", func() {
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: useAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes: []string{"BOGUS_DEVICE_GROUP_CHANGE"},
			},
		}
		group, err := apiDeviceGroup.ConvertToDBObject()
		Expect(group).ToNot(BeNil())
		Expect(err).Should(MatchError(ErrModelConversionWarning))
		Expect(fmt.Sprintln(err)).To(ContainSubstring("Unknown DeviceGroupChangeType 'BOGUS_DEVICE_GROUP_CHANGE'"))
	})

	// NOTE: Duplicate Changes tested in functional tests (requires live service)

	It("Ensure we can provide a timestamp for a livequery change", func() {
		yesterday := yesterday().UnixMilli()
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: useAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes:       []string{devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()},
				LiveQueryTime: &yesterday,
			},
		}
		deviceGroup, err := apiDeviceGroup.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(deviceGroup).ToNot(BeNil())
		Expect(deviceGroup.LiveQueryChangeTime).ToNot(BeNil())
		Expect(deviceGroup.LiveQueryChangeTime.UnixMilli()).To(BeEquivalentTo(yesterday))
	})

	It("Ensure we can provide a timestamp for a livequery change but not in the changes array", func() {
		yesterday := yesterday().UnixMilli()
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: useAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes:       []string{devicegroupchangetype.DG_POLICY_CHANGE.String()},
				LiveQueryTime: &yesterday,
			},
		}
		deviceGroup, err := apiDeviceGroup.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(deviceGroup).ToNot(BeNil())
		Expect(deviceGroup.LiveQueryChangeTime).To(BeNil())
		Expect(deviceGroup.LastPolicyChangeTime.UnixMilli()).To(BeNumerically("~", time.Now().UnixMilli(), 5))
	})

	It("Ensure livequery change with no timestamp uses current time", func() {
		apiDeviceGroup := APIDeviceGroup{
			APIDeviceGroupURI: useAPI,
			APIDeviceGroupBody: APIDeviceGroupBody{
				Changes: []string{devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()},
			},
		}
		deviceGroup, err := apiDeviceGroup.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(deviceGroup).ToNot(BeNil())
		Expect(deviceGroup.LiveQueryChangeTime).ToNot(BeNil())
		Expect(deviceGroup.LiveQueryChangeTime.UnixMilli()).To(BeNumerically("~", time.Now().UnixMilli(), 5))
	})
})
