// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

// Package models -- contains database models
package models

import (
	"fmt"
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	"gorm.io/gorm/clause"

	"gitlab.bit9.local/psc/ems/hinting/internal/utils"
	"go.uber.org/zap/zapcore"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicechangetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
)

////////////////////////////////////////////////////////////////////////////////
// Postgres Database Specifics

// Device -- device structure stored in postgres
//   - DeploymentType and DeviceType are legacy values and not currently stored
type Device struct {
	Id                          int64  `gorm:"primaryKey"`
	OrgKey                      string `gorm:"primaryKey"`
	GroupChangeTime             *time.Time
	SensorConfigUpdateTime      *time.Time
	SensorCertificateUpdateTime *time.Time
	HashDeleteRequestTime       *time.Time
	HashUploadRequestTime       *time.Time
	LogsRequestTime             *time.Time
	SensorKitUpdateTime         *time.Time
	PerfmonTriggerTime          *time.Time
	SensorActionTime            *time.Time
	ReRegisterTime              *time.Time
	LiveQueryChangeTime         *time.Time
	RepRequestTime              *time.Time
	DeploymentType              *deploymenttype.DeploymentType `gorm:"-:all" json:"-"` // We don't store these - it is provided when hints are requested
	DeviceType                  *devicetype.DeviceType         `gorm:"-:all" json:"-"`
	env                         string
}

func (device *Device) CacheKey() string {
	return fmt.Sprintf("devices-{%s}-%d", device.getOrgKey(), device.Id)
}

func (device *Device) getOrgKey() string {
	return device.OrgKey
}

func (device *Device) GetOnConflictClause() clause.OnConflict {
	return clause.OnConflict{
		Columns:   device.GetPrimaryColumnNames(),
		DoUpdates: device.GetUpsertClauses(),
	}
}

func (device *Device) GetPrimaryColumnNames() []clause.Column {
	return []clause.Column{{Name: "id"}, {Name: "org_key"}}
}

func (device *Device) GetUpsertClauses() clause.Set {
	gormExprMap := getGreatestExpr("devices", "group_change_time", "sensor_config_update_time",
		"sensor_certificate_update_time", "hash_delete_request_time", "hash_upload_request_time",
		"logs_request_time", "sensor_kit_update_time", "perfmon_trigger_time", "sensor_action_time", "re_register_time",
		"live_query_change_time", "rep_request_time")
	return clause.Assignments(gormExprMap)
}

func (device *Device) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	encoder.AddInt64("device_id", device.Id)
	encoder.AddString("org_key", device.OrgKey)
	utils.AddTimePointerToEncoder(encoder, "group_change_time", device.GroupChangeTime)
	utils.AddTimePointerToEncoder(encoder, "sensor_config_update_time", device.SensorConfigUpdateTime)
	utils.AddTimePointerToEncoder(encoder, "sensor_certificate_update_time", device.SensorCertificateUpdateTime)
	utils.AddTimePointerToEncoder(encoder, "hash_delete_request_time", device.HashDeleteRequestTime)
	utils.AddTimePointerToEncoder(encoder, "hash_upload_request_time", device.HashUploadRequestTime)
	utils.AddTimePointerToEncoder(encoder, "logs_request_time", device.LogsRequestTime)
	utils.AddTimePointerToEncoder(encoder, "sensor_kit_update_time", device.SensorKitUpdateTime)
	utils.AddTimePointerToEncoder(encoder, "perfmon_trigger_time", device.PerfmonTriggerTime)
	utils.AddTimePointerToEncoder(encoder, "sensor_action_time", device.SensorActionTime)
	utils.AddTimePointerToEncoder(encoder, "re_register_time", device.ReRegisterTime)
	utils.AddTimePointerToEncoder(encoder, "live_query_change_time", device.LiveQueryChangeTime)
	utils.AddTimePointerToEncoder(encoder, "rep_request_time", device.RepRequestTime)
	if device.DeploymentType != nil {
		encoder.AddString("deployment_type", string(*device.DeploymentType))
	}
	if device.DeviceType != nil {
		encoder.AddString("device_type", string(*device.DeviceType))
	}
	encoder.AddString("env", device.env)
	return nil
}

func (device *Device) GetMetricDimensions() *metrics.Dimensions {
	var hints []string
	dimensions := metrics.Dimensions{
		metrics.DimensionOrgKey:  device.OrgKey,
		metrics.DimensionEnvName: device.env,
	}

	if device.GroupChangeTime != nil {
		hints = append(hints, "dev_GroupChangeTime")
	}
	if device.SensorConfigUpdateTime != nil {
		hints = append(hints, "dev_SensorConfigUpdateTime")
	}
	if device.SensorCertificateUpdateTime != nil {
		hints = append(hints, "dev_SensorCertificateUpdateTime")
	}
	if device.HashDeleteRequestTime != nil {
		hints = append(hints, "dev_HashDeleteRequestTime")
	}
	if device.HashUploadRequestTime != nil {
		hints = append(hints, "dev_HashUploadRequestTime")
	}
	if device.LogsRequestTime != nil {
		hints = append(hints, "dev_LogsRequestTime")
	}
	if device.SensorKitUpdateTime != nil {
		hints = append(hints, "dev_SensorKitUpdateTime")
	}
	if device.PerfmonTriggerTime != nil {
		hints = append(hints, "dev_PerfmonTriggerTime")
	}
	if device.SensorActionTime != nil {
		hints = append(hints, "dev_SensorActionTime")
	}
	if device.ReRegisterTime != nil {
		hints = append(hints, "dev_ReRegisterTime")
	}
	if device.LiveQueryChangeTime != nil {
		hints = append(hints, "dev_LiveQueryChangeTime")
	}
	if device.RepRequestTime != nil {
		hints = append(hints, "dev_RepRequestTime")
	}

	if len(hints) == 0 {
		return nil
	}

	dimensions["dev_hints"] = strings.Join(hints, ",")

	return &dimensions
}

//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// API Input Handling

type APIDeviceURI struct {
	Id     int64  `uri:"deviceId" binding:"required"`
	OrgKey string `uri:"orgKey" binding:"required"`
}

// APIDeviceBody -- Definition of the body data
type APIDeviceBody struct {
	// List of device change types to be updated.
	//
	// Valid Change Types:
	// DEV_GROUP_CHANGE,
	// DEV_HASH_DELETE_REQUEST,
	// DEV_HASH_SAMPLE_REQUEST,
	// DEV_LIVE_QUERY_CHANGE
	// DEV_REREGISTER,
	// DEV_SENSOR_ACTION,
	// DEV_SENSOR_CERTIFICATE_REQUEST,
	// DEV_SENSOR_CONFIG_UPDATE,
	// DEV_SENSOR_KIT_UPDATE,
	// DEV_SENSOR_LOGS_REQUEST,
	// DEV_SENSOR_TRIGGER_PERFMON
	// DEV_REP_REQUEST
	// required: false
	// example: ["DEV_REREGISTER"]
	Changes []string `json:"changes" binding:"omitempty,max=12"`

	// If specified, overrides the current time value for any DEV_LIVE_QUERY_CHANGE.  Has no effect with other change types.
	//
	// required: false
	// example: 1672531200000
	LiveQueryTime *int64 `json:"live_query_time"`
}

type APIDevice struct {
	APIDeviceURI
	APIDeviceBody
}

func (apiDevice *APIDevice) GetURIPart() URIPart[*Device] {
	return &apiDevice.APIDeviceURI
}

func (apiDevice *APIDevice) GetBodyPart() any {
	return &apiDevice.APIDeviceBody
}

// ConvertToDBObject -- Creates a starting Device object with the URI information
func (uri *APIDeviceURI) ConvertToDBObject() (*Device, error) {
	uri.OrgKey = utils.Sanitize("orgkey", uri.OrgKey)
	_, env, err := utils.ValidateOrgKey(uri.OrgKey)
	if err != nil {
		return nil, err
	}
	return &Device{Id: uri.Id, OrgKey: uri.OrgKey, env: env}, nil
}

// ConvertToDBObject -- Translates the APIDevice structure obtained from a user/other service into a Device structure
// that is stored in the postgres database.
func (apiDevice *APIDevice) ConvertToDBObject() (device *Device, err error) {
	now := time.Now()
	device, err = apiDevice.APIDeviceURI.ConvertToDBObject()
	if err != nil {
		return nil, err
	}

	// NOTE: Additions/removals here will require updates in `test/functional/suites/lib/models/enumerations.py`
	//       as well as changes in `internal/app/models/enums/devicechangetype/devicechangetype.go`
	for _, changeString := range apiDevice.Changes {
		switch changeString {
		case devicechangetype.DEV_GROUP_CHANGE.String():
			device.GroupChangeTime = &now
		case devicechangetype.DEV_HASH_DELETE_REQUEST.String():
			device.HashDeleteRequestTime = &now
		case devicechangetype.DEV_HASH_SAMPLE_REQUEST.String():
			device.HashUploadRequestTime = &now
		case devicechangetype.DEV_LIVE_QUERY_CHANGE.String():
			if apiDevice.LiveQueryTime == nil {
				device.LiveQueryChangeTime = &now
				err = NewModelConversionWarning("Device DEV_LIVE_QUERY_CHANGE did NOT receive live_query_time")
			} else {
				device.LiveQueryChangeTime = utils.GetTimePointer(*apiDevice.LiveQueryTime)
			}
		case devicechangetype.DEV_REREGISTER.String():
			device.ReRegisterTime = &now
		case devicechangetype.DEV_SENSOR_ACTION.String():
			device.SensorActionTime = &now
		case devicechangetype.DEV_SENSOR_CERTIFICATE_REQUEST.String():
			device.SensorCertificateUpdateTime = &now
		case devicechangetype.DEV_SENSOR_CONFIG_UPDATE.String():
			device.SensorConfigUpdateTime = &now
		case devicechangetype.DEV_SENSOR_KIT_UPDATE.String():
			device.SensorKitUpdateTime = &now
		case devicechangetype.DEV_SENSOR_LOGS_REQUEST.String():
			device.LogsRequestTime = &now
		case devicechangetype.DEV_SENSOR_TRIGGER_PERFMON.String():
			device.PerfmonTriggerTime = &now
		case devicechangetype.DEV_REP_REQUEST.String():
			device.RepRequestTime = &now
		default:
			err = NewModelConversionWarning("Unknown DeviceChangeType: '%s'", changeString)
		}
	}

	return device, err
}

//
////////////////////////////////////////////////////////////////////////////////

// plusMinute -- If supplied time is not nil, return base time plus indicated minutes, otherwise return nil
func plusMinute(setTime *time.Time, mins int) *time.Time {
	if setTime == nil {
		return nil
	} else {
		updated := setTime.Add(time.Minute * time.Duration(mins))
		return &updated
	}
}

// NewDeviceAtTime -- Create a filled-in Device object with all time parameters set to the specified time.
// To assist in isolating incorrect assignments, the base time submitted will be ingremented by 1 minute for
// each entry.
func NewDeviceAtTime(id int64, orgKey string, deploymentType *deploymenttype.DeploymentType, deviceType *devicetype.DeviceType, setTime *time.Time) *Device {
	return &Device{
		Id:                          id,
		OrgKey:                      orgKey,
		GroupChangeTime:             setTime,
		SensorConfigUpdateTime:      setTime,
		SensorCertificateUpdateTime: setTime,
		HashDeleteRequestTime:       setTime,
		HashUploadRequestTime:       setTime,
		LogsRequestTime:             setTime,
		SensorKitUpdateTime:         setTime,
		SensorActionTime:            setTime,
		PerfmonTriggerTime:          setTime,
		ReRegisterTime:              setTime,
		LiveQueryChangeTime:         plusMinute(setTime, 0), // just used to ensure coverage
		DeviceType:                  deviceType,
		DeploymentType:              deploymentType,
		RepRequestTime:              setTime,
	}
}

// NewDeviceWindowsEndpoint -- Create a filled-in WINDOWS ENDPOINT Device object with all time parameters set to the
// specified time.
func NewDeviceWindowsEndpoint(id int64, orgkey string, setTime *time.Time) *Device {
	deviceType := devicetype.WINDOWS
	deploymentType := deploymenttype.ENDPOINT
	return NewDeviceAtTime(id, orgkey, &deploymentType, &deviceType, setTime)
}

// NewDeviceTimeVaries -- Create a filled-in Device object with all time parameters set to the base time plus
// additional minutes to make them different in order to assist in isolating incorrect assignments
func NewDeviceTimeVaries(id int64, orgKey string, setTime *time.Time) *Device {
	return &Device{
		Id:                          id,
		OrgKey:                      orgKey,
		GroupChangeTime:             plusMinute(setTime, 0),
		SensorConfigUpdateTime:      plusMinute(setTime, 1),
		SensorCertificateUpdateTime: plusMinute(setTime, 2),
		HashDeleteRequestTime:       plusMinute(setTime, 3),
		HashUploadRequestTime:       plusMinute(setTime, 4),
		LogsRequestTime:             plusMinute(setTime, 5),
		SensorKitUpdateTime:         plusMinute(setTime, 6),
		SensorActionTime:            plusMinute(setTime, 7),
		PerfmonTriggerTime:          plusMinute(setTime, 8),
		ReRegisterTime:              plusMinute(setTime, 9),
		LiveQueryChangeTime:         plusMinute(setTime, 10),
		RepRequestTime:              plusMinute(setTime, 11),
	}
}
