package models

import (
	"database/sql"
	"strings"
	"time"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/types"
	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
)

var _ = Describe("Postgres Organization Model tests", func() {
	It("results in nil dimensions with 'empty' organization", func() {
		d := &Organization{}

		m := d.GetMetricDimensions()

		Expect(m).To(BeNil())
	})

	It("results in nil dimensions if OrgKey, env, and killswitch are set", func() {
		d := &Organization{
			OrgKey: "ABC",
			env:    "local",
			IsKillSwitchOn: sql.NullBool{
				Bool:  false,
				Valid: false,
			},
		}

		m := d.GetMetricDimensions()

		Expect(m).To(BeNil())
	})

	It("properly sets dimensions with hints", func() {
		d := &Organization{
			OrgKey: "ABC",
			env:    "local",
			IsKillSwitchOn: sql.NullBool{
				Bool:  true,
				Valid: true,
			},
			ChangeTimestamps: types.OrgChangeset{
				orgchangetype.ORG_NET_CFG_CHANGE:    func(t time.Time) *time.Time { return &t }(time.Now()),
				orgchangetype.ORG_LIVE_QUERY_CHANGE: func(t time.Time) *time.Time { return &t }(time.Now()),
			},
		}

		m := *d.GetMetricDimensions()

		Expect(m).ToNot(BeNil())

		Expect(m[metrics.DimensionOrgKey]).To(Equal("ABC"))
		Expect(m[metrics.DimensionEnvName]).To(Equal("local"))
		Expect(m["is_killswitch_on"]).To(Equal("true"))
		Expect(strings.Contains(m["org_hints"], orgchangetype.ORG_NET_CFG_CHANGE.String())).To(BeTrue())
		Expect(strings.Contains(m["org_hints"], orgchangetype.ORG_LIVE_QUERY_CHANGE.String())).To(BeTrue())
		Expect(len(strings.Split(m["org_hints"], ","))).To(Equal(2))
		Expect(len(m)).To(Equal(4))
	})

	It("properly sets dimensions with deployment type hints", func() {
		d := &Organization{
			OrgKey: "ABC",
			env:    "local",
			IsKillSwitchOn: sql.NullBool{
				Bool:  true,
				Valid: true,
			},
			DeploymentTypeChangeTimestamps: types.DeploymentTypeToChangesMap{
				deploymenttype.AWS: types.OrgChangeset{
					orgchangetype.ORG_NET_CFG_CHANGE:    func(t time.Time) *time.Time { return &t }(time.Now()),
					orgchangetype.ORG_LIVE_QUERY_CHANGE: func(t time.Time) *time.Time { return &t }(time.Now()),
				},
			},
		}

		m := *d.GetMetricDimensions()

		Expect(m).ToNot(BeNil())

		Expect(m[metrics.DimensionOrgKey]).To(Equal("ABC"))
		Expect(m[metrics.DimensionEnvName]).To(Equal("local"))
		Expect(m["is_killswitch_on"]).To(Equal("true"))
		Expect(strings.Contains(m["AWS"], orgchangetype.ORG_NET_CFG_CHANGE.String())).To(BeTrue())
		Expect(strings.Contains(m["AWS"], orgchangetype.ORG_LIVE_QUERY_CHANGE.String())).To(BeTrue())
		Expect(len(strings.Split(m["AWS"], ","))).To(Equal(2))
		Expect(len(m)).To(Equal(4))
	})

	It("properly sets dimensions with canary payload hints", func() {
		d := &Organization{
			OrgKey: "ABC",
			env:    "local",
			IsKillSwitchOn: sql.NullBool{
				Bool:  true,
				Valid: true,
			},
			CanaryPayloadTimestamps: types.DeviceTypeToTimeMap{
				devicetype.WINDOWS: time.Now(),
				devicetype.MAC:     time.Now(),
			},
		}

		m := *d.GetMetricDimensions()

		Expect(m).ToNot(BeNil())

		Expect(m[metrics.DimensionOrgKey]).To(Equal("ABC"))
		Expect(m[metrics.DimensionEnvName]).To(Equal("local"))
		Expect(m["is_killswitch_on"]).To(Equal("true"))
		Expect(strings.Contains(m["org_canary_payload"], string(devicetype.WINDOWS))).To(BeTrue())
		Expect(strings.Contains(m["org_canary_payload"], string(devicetype.MAC))).To(BeTrue())
		Expect(len(strings.Split(m["org_canary_payload"], ","))).To(Equal(2))
		Expect(len(m)).To(Equal(4))
	})
})
