// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package app

import (
	"context"
	"errors"
	"fmt"
	"net"
	"net/http"
	"sync/atomic"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/controllers"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/middleware"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	"github.com/gin-contrib/cors"
	"github.com/gin-gonic/gin"
	"gitlab.bit9.local/psckit/pscrbac"
	"go.uber.org/zap"
)

// ServiceOk -- Hinting service is available (and ready, if isReady is queried as true)
// swagger:response serviceOk
type ServiceOk http.Response

// ServiceUnavailable -- Hinting service not available, or is not ready
// swagger:response serviceUnavailable
type ServiceUnavailable http.Response

// OKResponse -- Operation successfully completed
// swagger:response okResponse
type OKResponse http.Response

// BadRequestResponse -- Operation unsuccessful due to a malformed request body or incompatible data
// swagger:response badRequestResponse
type BadRequestResponse http.Response

// UnauthorizedResponse -- Operation unsuccessful due to an authentication failure
// swagger:response unauthorizedResponse
type UnauthorizedResponse http.Response

// NotFoundResponse -- Operation failed due to missing URI parameters
// swagger:response notFoundResponse
type NotFoundResponse http.Response

func SetupRouter(service *HintingService, mwf *middleware.RouteMiddlewareFactory) *gin.Engine {
	logger.BaseLogger().Debug("Setting up router...")

	router := gin.New()

	if service.cfg.IsLocalEnv() {
		// Enable CORS (cross-origin resource sharing) support for swagger.
		router.Use(cors.New(cors.Config{
			AllowOrigins:     []string{"https://editor.swagger.io", "https://editor-next.swagger.io"},
			AllowMethods:     []string{"GET, POST, PUT, DELETE, OPTIONS"},
			AllowHeaders:     []string{"Content-Type,access-control-allow-origin, access-control-allow-headers, access-control-allow-methods"},
			ExposeHeaders:    []string{"Content-Length"},
			AllowCredentials: true,
		}))
	}

	// --------------------------------------------------------------------------------

	// swagger:route GET /healthCheck Service HealthCheck
	//
	// Health check API for the hinting microservice.
	//
	// Used by the kubernetes cluster to determine if the service is active and ready.
	//
	// Parameters:
	//   + name: isReady
	//     in: query
	//     description: If supplied and true, check for readiness instead of simply checking for aliveness
	//     required: false
	//     type: boolean
	//
	//     Responses:
	//       200: serviceOk
	//       503: serviceUnavailable
	router.GET("/hinting/healthCheck", func(ctx *gin.Context) {
		isReadyCheck := ctx.Param("isReady")
		if isReadyCheck == "true" {
			if atomic.LoadInt32(service.Ready) == 1 {
				ctx.Status(http.StatusOK)
				return
			} else {
				ctx.Status(http.StatusServiceUnavailable)
				return
			}
		}
		if atomic.LoadInt32(service.alive) == 1 {
			ctx.Status(http.StatusOK)
			return
		}
		ctx.Status(http.StatusServiceUnavailable)
	})

	// --------------------------------------------------------------------------------

	v1 := router.Group("/hinting/v1/")
	v1.Use(mwf.TranslateGinContext())

	db := service.dbClient
	redisClient := service.redisClient
	metricsWrapper := metrics.NewMetricsClientWrapper(service.metricsClient)

	deviceController := controllers.NewController[*models.Device, models.APIDevice, *models.APIDevice](db, redisClient, metricsWrapper)
	// swagger:route PUT /v1/orgs/{orgKey}/devices/{deviceId} Update V1UpdateDevice
	//
	// Create/Update device timestamps.
	//
	// Updates the associated timestamps for a specific device within a specific organization.
	// Any changes indicated will have their timestamps updated to the UTC time of the hinting service.
	//
	//	Consumes:
	//	- application/json
	//
	//	Produces:
	//	- application/json
	//
	//  Responses:
	//       200: okResponse
	//       400: badRequestResponse
	//       401: unauthorizedResponse
	v1.PUT("orgs/:orgKey/devices/:deviceId", mwf.GetHandlersForRBACRoute("HTTP PUT device", deviceController.Update, pscrbac.NewPermission("csr.hints.device", pscrbac.Create))...)

	deviceGroupController := controllers.NewController[*models.DeviceGroup, models.APIDeviceGroup, *models.APIDeviceGroup](db, redisClient, metricsWrapper)
	// swagger:route PUT /v1/orgs/{orgKey}/devicegroups/{deviceGroupId} Update V1UpdateDeviceGroup
	//
	// Create/Update device group timestamps.
	//
	// Updates the associated timestamps for a specific device group within a specific organization.
	// Any changes indicated will have their timestamps updated to the UTC time of the hinting service.
	//
	//	Consumes:
	//	- application/json
	//
	//	Produces:
	//	- application/json
	//
	//     Responses:
	//       200: okResponse
	//       400: badRequestResponse
	//       401: unauthorizedResponse
	v1.PUT("orgs/:orgKey/devicegroups/:deviceGroupId", mwf.GetHandlersForRBACRoute("HTTP PUT devicegroup", deviceGroupController.Update, pscrbac.NewPermission("csr.hints.policy", pscrbac.Create))...)

	orgController := controllers.NewController[*models.Organization, models.APIOrg, *models.APIOrg](db, redisClient, metricsWrapper)
	// swagger:route PUT /v1/orgs/{orgKey} Update V1UpdateOrg
	//
	// Create/Update organization timestamps.
	//
	// Updates the associated timestamps for a specific organization within a defined environment.
	// Any changes indicated will have their timestamps updated to the UTC time of the hinting service.
	//
	//	Consumes:
	//	- application/json
	//
	//	Produces:
	//	- application/json
	//
	//     Responses:
	//       200: okResponse
	//       400: badRequestResponse
	//       401: unauthorizedResponse
	v1.PUT("orgs/:orgKey", mwf.GetHandlersForRBACRoute("HTTP PUT org", orgController.Update, pscrbac.NewPermission("csr.hints.org", pscrbac.Create))...)

	envController := controllers.NewController[*models.Environment, models.APIEnv, *models.APIEnv](db, redisClient, metricsWrapper)
	// swagger:route PUT /v1/envs/{envName} Update V1UpdateEnv
	//
	// Create/Update environment timestamps.
	//
	// Updates the associated timestamps for a specific environment.
	// Any changes indicated will have their timestamps updated to the UTC time of the hinting service.
	//
	//	Consumes:
	//	- application/json
	//
	//	Produces:
	//	- application/json
	//
	//     Responses:
	//       200: okResponse
	//       400: badRequestResponse
	//       401: unauthorizedResponse
	v1.PUT("envs/:envName", mwf.GetHandlersForRBACRoute("HTTP PUT env", envController.Update, pscrbac.NewPermission("csr.hints.env", pscrbac.Create))...)

	// --------------------------------------------------------------------------------

	hintController := controllers.NewHintControllerFromDbWithCache(db, redisClient, metricsWrapper)
	// swagger:route POST /v1/hints GetHints V1PostHints
	//
	// Returns a list of hints and any updated timestamps.
	//
	// GetHints returns the list of hints for a specified device, device group and org in an env.
	//
	//	Consumes:
	//	- application/json
	//
	//	Produces:
	//	- application/json
	//
	//  responses:
	//    200: hintResult
	//    400: badRequestResponse
	//    401: unauthorizedResponse
	v1.POST("hints", mwf.GetHandlersForRoute("HTTP POST hints", hintController.GetHints)...)

	logger.BaseLogger().Debug("Finished setting up router...")
	return router
}

type HttpServer struct {
	ginEngine  *gin.Engine
	httpServer *http.Server
	Port       string
}

func NewHttpServer(service *HintingService) (*HttpServer, error) {
	mwf := middleware.NewMiddlewareFactory(service.Logger, service.metricsClient, service.cfg)
	router := SetupRouter(service, mwf)

	if service.cfg.ProfilingEnabled {
		middleware.RouteProfiler(router, service.cfg)
	}

	httpServer := &http.Server{
		Addr: service.cfg.ServerAddr,
		Handler: mwf.ApplyTracingMiddleware(
			mwf.GetSpanErrorTaggingMiddleware(
				mwf.Logging(
					mwf.WriteStatus(
						mwf.TracingRequestIdAndLogger(
							mwf.GetRecoveryMiddleware(
								mwf.TimeoutMiddleware(router, service.cfg.HttpHandleTimeout).ServeHTTP,
							)))))),
		ReadHeaderTimeout: service.cfg.HttpReadTimeout,
		ReadTimeout:       service.cfg.HttpReadTimeout,
		WriteTimeout:      service.cfg.HttpWriteTimeout,
	}
	listener, err := net.Listen("tcp", service.cfg.ServerAddr)
	if err != nil {
		return nil, err
	}
	go func() {
		if err := httpServer.Serve(listener); err != nil && !errors.Is(err, http.ErrServerClosed) {
			logger.BaseLogger().Error("listen: %s\n", zap.Error(err))
		}
	}()
	return &HttpServer{ginEngine: router, httpServer: httpServer, Port: fmt.Sprintf("%d", listener.Addr().(*net.TCPAddr).Port)}, nil
}

func (server *HttpServer) Shutdown() error {
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	if err := server.httpServer.Shutdown(ctx); err != nil {
		return err
	}
	return nil
}

// //////////////////////////////////////////////////////////////////////////////
// Swagger definitions for responses and requests (placed here so as to be
// properly picked up by go-swagger)
// //////////////////////////////////////////////////////////////////////////////

// HintResult containing the list of hints, lastUpdateTime and liveQueryTime
// swagger:response hintResult
type HintResult struct {
	// in: body
	Body struct {
		// type: array
		// items:
		//   type: string
		// example: ["NEW_POLICY", "LIVE_QUERY"]
		Hints             []string `json:"hints"`
		LastLiveQueryTime int64    `json:"last_live_query_time,omitempty"`
		LastUpdateTime    int64    `json:"last_update_time,omitempty"`
	}
}

// swagger:parameters V1UpdateDevice
type V1UpdateDevice struct {
	// Unique organization identifier within an environment.
	//
	// in: path
	// required: true
	// example: 7ZP4H5XR
	OrgKey string `json:"orgKey" binding:"required"`

	// Unique device identifier.
	//
	// in: path
	// required: true
	// example: 123
	Id int64 `json:"deviceId" binding:"required"`

	// Request body with change types and device information.
	//
	// in: body
	// required: true
	Body models.APIDeviceBody
}

// swagger:parameters V1UpdateDeviceGroup
type V1UpdateDeviceGroup struct {
	// Unique organization identifier within an environment.
	//
	// in: path
	// required: true
	// example: 7ZP4H5XR
	OrgKey string `json:"orgKey" binding:"required"`

	// Unique devicegroup identifier.
	//
	// in: path
	// required: true
	// example: 456
	Id string `json:"deviceGroupId" binding:"required"`

	// Request body with change types.
	//
	// in: body
	// required: true
	Body models.APIDeviceGroupBody
}

// swagger:parameters V1UpdateOrg
type V1UpdateOrg struct {
	// Unique organization identifier within an environment.
	//
	// in: path
	// required: true
	// example: 7ZP4H5XR
	OrgKey string `json:"orgKey" binding:"required"`

	// Request body with change types.
	//
	// in: body
	// required: true
	Body models.APIOrgBody
}

// swagger:parameters V1UpdateEnv
type V1UpdateEnv struct {
	// Unique identifier for an environment.
	//
	// in: path
	// required: true
	// example: local
	Name string `json:"envName" binding:"required"`

	// Request body with change types.
	//
	// in: body
	// required: true
	Body models.APIEnvBody
}

// swagger:parameters V1PostHints
type V1PostHints struct {
	// Request body with device information
	//
	// in: body
	// required: true
	Body hints.JsonHintRequest
}
