// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package dbwrapper -- wrapper for the postgres database
package dbwrapper

import (
	"fmt"
	"strings"

	"go.uber.org/zap"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	"gorm.io/driver/postgres"
	"gorm.io/gorm"
	"gorm.io/plugin/dbresolver"
)

type DbClient struct {
	*gorm.DB
	DbHost string
}

func NewDbClient(cfg *config.PostgresqlConfig) (*DbClient, error) {
	db, err := ConnectToDB("Hinting", cfg)
	if err != nil {
		return nil, fmt.Errorf("error connecting to database %v", err)
	}
	client := &DbClient{DB: db, DbHost: cfg.DbHost}
	err = client.Start()
	return client, err
}

func NewDbClientWithMetrics(cfg *config.Config, alive, ready *int32, metricsClient metrics.Service) (*DbClient, error) {
	dbClient, err := NewDbClient(&cfg.PostgresqlConfig)
	if err != nil {
		logger.BaseLogger().Error("Unable to get new database client", zap.Error(err))
	} else {
		underlyingDb, err2 := dbClient.DB.DB()
		if err2 != nil {
			logger.BaseLogger().Error("Unable to get underlying db", zap.Error(err2))
			return nil, err2
		}
		if metricsClient != nil {
			RegisterDBStatMetrics(underlyingDb, metricsClient, ready, alive, cfg)
		}
	}
	return dbClient, err
}

func (dbc *DbClient) Start() error {
	if err := waitForDbToBeReady(100, dbc.DB); err != nil {
		return err
	}
	return nil
}

func (dbc *DbClient) Close() error {
	sqlDB, err := dbc.DB.DB()
	if err != nil {
		return err
	} else {
		if err := sqlDB.Close(); err != nil {
			return err
		}
	}
	return nil
}

func waitForDbToBeReady(retries int, db *gorm.DB) error {
	var err error = nil
	for i := 0; i < retries; i++ {
		err = db.Exec("SELECT 1").Error
		if err == nil {
			return nil
		}
	}
	return err
}

func ConnectToDB(description string, postgresqlConfig *config.PostgresqlConfig) (db *gorm.DB, err error) {
	db, err = gorm.Open(getDialector(*postgresqlConfig, false))
	if err == nil {
		logger.BaseLogger().Debug(fmt.Sprintf("Successfully connected to postgres database (%s).", description))
	} else {
		return nil, err
	}

	// Enable use of the reader core
	if postgresqlConfig.DbHostReader != "" && postgresqlConfig.DbHostReader != postgresqlConfig.DbHost {
		logger.BaseLogger().Info("Enabling DB Reader")

		err = db.Use(dbresolver.Register(dbresolver.Config{
			Replicas: []gorm.Dialector{getDialector(*postgresqlConfig, true)},
		}))
		if err != nil {
			logger.BaseLogger().Error("Unable to configure DB reader core")
			return nil, err
		}
	} else {
		logger.BaseLogger().Warn("No DB Reader Endpoint specified")
	}

	underlyingDb, err := db.DB()
	if err != nil {
		logger.BaseLogger().Error("Unable to get underlying db to configure conn pool settings!")
		return nil, err
	}
	underlyingDb.SetConnMaxLifetime(postgresqlConfig.SQLMaxConnLifetimeMinutes)
	underlyingDb.SetMaxIdleConns(postgresqlConfig.SQLMaxIdleConns)
	underlyingDb.SetMaxOpenConns(postgresqlConfig.SQLMaxOpenConns)
	return
}

func getDatabaseConnectionStringFromConfig(c config.PostgresqlConfig, useReader bool) string {
	var builder strings.Builder
	configParams := map[string]string{
		"port":        c.DbPort,
		"user":        c.DbUser,
		"password":    c.DbPass,
		"dbname":      c.DbName,
		"sslmode":     c.PgSSLMode,
		"sslrootcert": c.SSLRootCert,
	}

	if useReader {
		configParams["host"] = c.DbHostReader
	} else {
		configParams["host"] = c.DbHost
	}

	for key, value := range configParams {
		if value != "" {
			builder.WriteString(fmt.Sprintf("%s=%s ", key, value))
		}
	}
	return strings.TrimSpace(builder.String())
}

func getDialector(postgresqlConfig config.PostgresqlConfig, useReader bool) gorm.Dialector {
	if postgresqlConfig.TestDatabaseConn == nil {
		return postgres.Open(getDatabaseConnectionStringFromConfig(postgresqlConfig, useReader))
	} else {
		return postgresqlConfig.TestDatabaseConn
	}
}
